import asyncio
import multiprocessing

from decimal import Decimal
from typing import Dict, Optional, List

from pyqqq.brokerage.tracker import TradingTracker
from pyqqq.brokerage.kis.oauth import KISAuth
from pyqqq.brokerage.kis.simple import KISSimpleDomesticStock
from pyqqq.brokerage.ebest.oauth import EBestAuth
from pyqqq.brokerage.ebest.simple import EBestSimpleDomesticStock
from pyqqq.datatypes import *
from pyqqq.utils.logger import get_logger, logging

class TradingTrackerMultiProcessController:
    """
        부모 프로세스에서 생성되는 TradingTracker Controller
        메시지 큐를 통해 자식 프로세스에서 돌아가는 TradingTracker를 제어한다.
        자식 프로세스에서 listen한 order 이벤트는 event_snapshot_queue를 통해 받을 수 있다.
    """

    def __init__(
        self,
        auth_info: Dict[str, str] = None,
        fee_rate: Decimal = Decimal(0.00015),  # 뱅키스, LS증권 수수료율 0.015%
    ):

        self.auth_info = auth_info
        self.fee_rate = fee_rate
        self.event_snapshot_queue = multiprocessing.Queue()
        self.message_queue = multiprocessing.Queue()

    def start(self):
        print("process controller. start called")
        self.process = multiprocessing.Process(target=run_subprocess, args=(self.fee_rate, self.event_snapshot_queue, self.message_queue, self.auth_info))
        self.process.start()
    
    def stop(self):
        print("process controller. stop called")
        if self.process is not None:
            self.message_queue.put("stop")
            self.process.join()

            print("process controller. process joined")

    def get_event_snapshot(self):
        if not self.event_snapshot_queue.empty():
            return self.event_snapshot_queue.get()
        return None

class TradingTrackerMultiProcessWorker:
    """
        자식 프로세스에서만 생성되는 TradingTracker Worker     

    Args:
        event_snapshot_queue (multiprocessing.Queue): 거래 이벤트가 발생시, 부모 프로세스에 스냅샷을 전송하는 큐
        message_queue (multiprocessing.Queue): 부모 프로세스에서 명령을 전달할때 사용하는 큐
        fee_rate (Decimal): 증권사 수수료율 (기본값: 0.015%)
    """

    logger = get_logger("TradingTrackerMultiProcessWorker", level=logging.INFO)

    def __init__(
        self,
        event_snapshot_queue: multiprocessing.Queue,
        message_queue: multiprocessing.Queue,
        auth_info: Dict[str, str] = {},
        fee_rate: Decimal = Decimal(0.00015),  # 뱅키스, LS증권 수수료율 0.015%
    ):
        print("TradingTrackerMultiProcessWorker init")
        self.event_snapshot_queue = event_snapshot_queue
        self.message_queue = message_queue
        self.stop_event = asyncio.Event()

        # auth_info를 외부에서 넣어주는 경우
        if auth_info is not None:
            if "KIS_APP_KEY" in auth_info:
                self.auth = KISAuth(
                    appkey=auth_info["KIS_APP_KEY"],
                    appsecret=auth_info["KIS_APP_SECRET"],
                    paper_trading=auth_info["PAPER_TRADING"] == "1",
                )

                self.simple_api = KISSimpleDomesticStock(
                    auth=self.auth,
                    account_no=auth_info["KIS_CANO"],
                    account_product_code=auth_info["KIS_ACNT_PRDT_CD"],
                    hts_id=auth_info["KIS_HTS_ID"],
                )
            elif "EBEST_APP_KEY" in auth_info:
                self.auth = EBestAuth(
                    auth_info["EBEST_APP_KEY"],
                    auth_info["EBEST_APP_SECRET"],
                    paper_trading=auth_info["PAPER_TRADING"] == "1",
                )
                self.simple_api = EBestSimpleDomesticStock(auth=self.auth)

        else:
            if os.getenv("EBEST_APP_KEY") is not None and os.getenv("KIS_APP_KEY") is None:
                self.auth = EBestAuth(
                    os.getenv("EBEST_APP_KEY"),
                    os.getenv("EBEST_APP_SECRET"),
                    paper_trading=os.getenv("PAPER_TRADING") == "1",
                )
                self.simple_api = EBestSimpleDomesticStock(auth=self.auth)

            else:
                self.auth = KISAuth(
                    appkey=os.getenv("KIS_APP_KEY"),
                    appsecret=os.getenv("KIS_APP_SECRET"),
                    paper_trading=os.getenv("PAPER_TRADING") == "1",
                )

                self.simple_api = KISSimpleDomesticStock(
                    auth=self.auth,
                    account_no=os.getenv("KIS_CANO"),
                    account_product_code=os.getenv("KIS_ACNT_PRDT_CD"),
                    hts_id=os.getenv("KIS_HTS_ID"),
                )
        self.tracker = TradingTracker(self.simple_api, fee_rate=fee_rate)
        self.tracker.on_pending_order_update = self.on_pending_order_update

    async def start(self):
        """
        Tracker와 메시지 큐 리스너를 시작합니다
        """
        self.tasks = [
            asyncio.create_task(self.tracker.start()),
            asyncio.create_task(self._listen_message_queue()),
        ]
        await asyncio.gather(*self.tasks)

    async def stop(self):
        """
        Tracker와 메시지 큐 리스터를 중단합니다.
        """
        self.stop_event.set()
        await self.tracker.stop()
        for t in self.tasks:
            t.cancel()

        await asyncio.gather(*self.tasks)
        

    async def _listen_message_queue(self):
        """
        이벤트 루프에서 실행되는 함수. 외부에서 메시지 큐에 넣어준 명령을 수행한다.
        """
        self.logger.info(f"start _listen_message_queue")
        while not self.stop_event.is_set():
            if not self.message_queue.empty():
                message = self.message_queue.get()
                if message == "stop":
                    await self.stop()
                    return

            await asyncio.sleep(1)
        self.logger.info(f"end _listen_message_queue")


    def on_pending_order_update(
        self,
        status,
        event: OrderEvent,
    ):
        self.event_snapshot_queue.put(self._make_event_snapshot(status, event))

    def _make_event_snapshot(self, status, event: OrderEvent):
        """
        OrderEvent를 이벤트 스냅샷으로 변환합니다
        """
        
        return {
            "account_info": self.simple_api.get_account(),
            "pending_orders": self.tracker.pending_orders,
            "positions": self.tracker.positions,
            "status": status,
            "event": event,
        }

def run_subprocess(fee_rate, event_snapshot_queue, message_queue, auth_info=None):
    """
    args
    event_snapshot_queue: multiprocessing.Queue[Dict]
        거래 이벤트와 그 상태의 position, order, 계좌정보가 담겨있는 딕셔너리
    message_queue: multiprocessing.Queue[String]
        TradingTrackerMultiProcessWorker를 제어하는데 사용되는 메시지 큐
        
    """
    
    print(f"run_subprocess called")
    worker = TradingTrackerMultiProcessWorker(fee_rate=fee_rate, event_snapshot_queue=event_snapshot_queue, message_queue=message_queue, auth_info=auth_info)
    asyncio.run(worker.start())


import os
import functools

from diskcache import Cache


class DiskCacheManager:
    def __init__(self, cache_name):
        parent_dir = os.getenv("CACHE_DIR") or "cache/"
        dir_str = os.path.join(parent_dir, cache_name)
        self.cache = Cache(dir_str)
        self.use_cache = os.getenv("STRATEGY_NAME") is None

    def memoize(self, expire=None):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if not self.use_cache:
                    return func(*args, **kwargs)
                if expire is None:
                    return self.cache.memoize()(func)(*args, **kwargs)
                else:
                    return self.cache.memoize(expire=expire)(func)(*args, **kwargs)
            return wrapper
        return decorator



import os
import pickle
import time
from datetime import datetime
from typing import List
from dataclasses import dataclass

import pika
from pyqqq.datatypes import *
from pyqqq.utils.logger import get_logger


class RabbitmqPubSub:
    """
    RabbitMQ 의 Publish/Subscribe 패턴을 사용하기 위해 연결 후 채널을 생성, 관리하는 클래스이다.
    Publisher는 fanout type으로 exchange에 메시지를 브로드캐스팅하고, Subscriber는 해당 exchange에 바인딩된 각각의 queue에서 메시지를 받는다.
    """
    logger = get_logger(f'{__name__}.RabbitmqPubSub')

    def __init__(self, host=None):
        self.host = host
        self.connect()

    def __del__(self):
        # if self.exchange_name:
        #    self.delete_exchange()
        try:
            if self.rabbitmq and not self.rabbitmq.is_closed:
                self.rabbitmq.close()
        except Exception:
            pass

    def connect(self):
        while True:
            try:
                params = pika.ConnectionParameters(
                    host=self.host or os.getenv('RABBITMQ_HOST') or 'localhost',
                    heartbeat=600,
                    blocked_connection_timeout=300,
                )
                self.rabbitmq = pika.BlockingConnection(params)
                self.channel = self.rabbitmq.channel()
                return
            except Exception as e:
                self.logger.warning(f'Reconnecting to rabbitmq... : {e}')
                time.sleep(1)

    def declare_exchange(self, exchange_name):
        if not exchange_name:
            raise ValueError("Exchange name is required")

        self.exchange_name = exchange_name
        self.channel.exchange_declare(exchange=exchange_name, exchange_type='fanout') # fanout 은 broadcast 방식. 즉, 모든 큐에 메시지를 전달한다.

    def delete_exchange(self):
        if self.exchange_name:
            self.channel.exchange_delete(exchange=self.exchange_name)

    def declare_queue(self, queue_name):
        if not self.exchange_name:
            raise ValueError("Exchange name is not declared yet")

        result = self.channel.queue_declare(queue=queue_name, exclusive=False) # exclusive=True이면 종료 시 큐가 삭제되어 메시지를 받을 수 없음
        return result.method.queue

    def bind_queue(self, queue_name):
        if not self.exchange_name:
            raise ValueError("Exchange name is not declared yet")

        self.channel.queue_bind(exchange=self.exchange_name, queue=queue_name)

    def unbind_queue(self, queue_name):
        if not self.exchange_name:
            raise ValueError("Exchange name is not declared yet")

        self.channel.queue_unbind(exchange=self.exchange_name, queue=queue_name)

    def delete_queue(self, queue_name):
        self.channel.queue_delete(queue=queue_name)

    def publish(self, message: str):
        """
        메시지를 발행한다.
        """
        if not self.rabbitmq or self.rabbitmq.is_closed:
            self.logger.warning('reconnecting...')
            self.connect()

        if not self.exchange_name:
            raise ValueError("Exchange name is not declared yet")

        try:
            self.channel.basic_publish(exchange=self.exchange_name, routing_key='', body=message)
        except pika.exceptions.StreamLostError:
            self.logger.warning('reconnect while publish...')
            time.sleep(1)
            self.rabbitmq = None
            self.publish(message=message)
        self.logger.info(f" [x] Sent : {message}")

    def run_subscribe(self, queue_name, on_message_callback):
        """
        메시지 구독을 시작한다. (CTRL+C로 종료)
        """
        if not self.rabbitmq or self.rabbitmq.is_closed:
            self.logger.warning('reconnecting...')
            self.connect()

        self.channel.basic_consume(queue=queue_name, on_message_callback=on_message_callback, auto_ack=False) # auto_ack=True이면, 에러 등의 상황으로 프로그램 종료 시 이후 메시지들이 소실될 수 있음
        self.logger.info(f' [*] Waiting for messages in {queue_name}. To exit press CTRL+C')
        try:
            self.channel.start_consuming()
        except pika.exceptions.StreamLostError:
            self.logger.warning('reconnect while subscribing...')
            time.sleep(1)
            self.run_subscribe(queue_name, on_message_callback)
        except KeyboardInterrupt as e:
            self.channel.stop_consuming()
            self.logger.info(" [x] Stopped consuming")
            raise e


class CopyCatPublisher(RabbitmqPubSub):
    """
    메시지를 다수의 Subscriber가 받을 수 있도록 브로드캐스팅하는 Publisher이다.

    TODO:
        - [ ] 사용자 권한 설정 ex) subscriber는 exchange에 write할 수 없어야 함
        - [ ] exchange와 queue의 이름의 uniqueness를 보장해야 함
        - [ ] 수정 및 취소 건은 org_order_no 가 서로 일치하지 않으므로 다른 방법으로 따라하도록 수정해야 함
    """
    logger = get_logger(f'{__name__}.CopyCatPublisher')

    def __del__(self):
        # self.destroy()
        super().__del__()

    @classmethod
    def build(cls, exchange_name, queue_names=[], host=None):
        """
        하나의 주제(exchange_name)로 여러 개의 메시지 큐를 포함한 Publisher를 생성한다.
        각각의 메시지 큐는 독립적으로 메시지를 받아서 처리할 수 있다.
        """

        c = cls(host=host)
        c.declare_exchange(exchange_name)
        for q in queue_names:
            c.declare_queue(q)
        return c

    def destroy(self, queue_names=[]):
        for q in queue_names:
            self.unbind_queue(q)
            self.delete_queue(q)
        self.delete_exchange()

    def add_mq(self, queue_name):
        """
        Publisher에 새로운 메시지 큐를 추가한다.

        TODO: 이름 중복일 때?
        """

        self.declare_queue(queue_name)


class CopyCatSubscriber(RabbitmqPubSub):
    """
    Publisher가 발행한 이벤트 메시지를 받아서 순차적으로 모방할 수 있도록 도와주는 카피캣이다.
    """
    logger = get_logger(f'{__name__}.CopyCatSubscriber')

    def __init__(self, queue_name, host=None):
        super().__init__(host=host)
        self.queue_name = queue_name

    def __del__(self):
        # self.destroy()
        super().__del__()

    @classmethod
    def build(cls, exchange_name, queue_name, host=None):
        c = cls(queue_name, host=host)
        c.declare_exchange(exchange_name)
        c.bind_queue(queue_name)
        return c

    def destroy(self):
        if self.queue_name:
            self.unbind_queue(self.queue_name)
            self.delete_queue(self.queue_name)

    def run(self, on_message_callback=None):
        # on_message_callback 이 함수가 아니면 self._on_message 함수를 사용
        if not on_message_callback or not callable(on_message_callback):
            on_message_callback = self._on_message
        self.run_subscribe(self.queue_name, on_message_callback)

    def _on_message(self, ch, method, properties, body):
        ch.basic_ack(delivery_tag=method.delivery_tag)

        message = pickle.loads(body)
        dt = message['dt']
        command = message['command']
        self.logger.info(f" [x] Received {command} at {dt.isoformat()}")
        data = message.get('order_form')
        if command == OrderRequestType.NEW or command == MessageType.OrderRequestNew:
            self.logger.info(f" [x] New order: {data}")
            # TODO: 따라하기
        elif command == OrderRequestType.MODIFY or command == MessageType.OrderRequestModify:
            self.logger.info(f" [x] Modify order: {data}")
            # TODO: 어떤 주문에서 내용을 수정할 것인지 알아내는 것은 쉽지 않다.
        elif command == OrderRequestType.CANCEL or command == MessageType.OrderRequestCancel:
            self.logger.info(f" [x] Cancel order: {data}")
            # TODO: 어떤 주문에서 취소할 것인지 알아내는 것은 쉽지 않다.
        elif command == MessageType.PositionsReport:
            data = message.get('positions', [])
            self.logger.info(f" [x] Total positions: {len(data)}")
            for position in data:
                self.logger.info(f" [x] {position}")
        else:
            self.logger.info(f" [x] Unknown command: {command}")


@dataclass
class OrderForm:
    asset_code: str
    status: TransactionStatus = None # 주문/체결 구분자
    side: OrderSide = None
    order_type: OrderType = None
    price: int = None
    quantity: int = None


@dataclass
class SimplePosition:
    asset_code: str
    asset_name: str
    quantity: int


class MessageType(Enum):
    OrderRequestNew = 1
    OrderRequestModify = 2
    OrderRequestCancel = 3
    PositionsReport = 4


class CopyCatMessage:
    @staticmethod
    def make_for_create(asset_code: str, status: TransactionStatus, side: OrderSide, order_type: OrderType, price: int = None, quantity: int = None):
        command = MessageType.OrderRequestNew
        status = status or TransactionStatus.ORDER
        order_form = OrderForm(asset_code=asset_code, status=status, side=side, order_type=order_type, price=price, quantity=quantity)
        return pickle.dumps({
            "dt": datetime.now(),
            "command": command,
            "order_form": order_form,
        })

    @staticmethod
    def make_for_update(asset_code: str, order_type: OrderType, price: int = None, quantity: int = None):
        command = MessageType.OrderRequestModify
        order_form = OrderForm(asset_code=asset_code, order_type=order_type, price=price, quantity=quantity)
        return pickle.dumps({
            "dt": datetime.now(),
            "command": command,
            "order_form": order_form,
        })

    @staticmethod
    def make_for_cancel(asset_code: str, quantity: int = None):
        command = MessageType.OrderRequestCancel
        order_form = OrderForm(asset_code=asset_code, quantity=quantity)
        return pickle.dumps({
            "dt": datetime.now(),
            "command": command,
            "order_form": order_form,
        })

    @staticmethod
    def make_for_positions(positions: List[StockPosition]):
        command = MessageType.PositionsReport
        results = []
        for position in positions:
            results.append(SimplePosition(asset_code=position.asset_code, asset_name=position.asset_name, quantity=position.quantity))
        return pickle.dumps({
            "dt": datetime.now(),
            "command": command,
            "positions": results,
        })

    @staticmethod
    def parse(message: str):
        return pickle.loads(message)



from io import BytesIO
from pyqqq.utils.api_client import send_request
from pyqqq.utils.logger import get_logger
import datetime as dtm
import pandas as pd
import pyqqq.config as c

logger = get_logger("pyqqq.data.daily")


def get_tick_data(date: dtm.date, asset_code: str) -> pd.DataFrame:
    """
    해당 일 체결 정보를 반환합니다

    Args:
        date (dtm.date): 조회할 날짜
        asset_code (str): 조회할 종목 코드

    Returns:
        pd.DataFrame: 요청한 일자의 해당 종목의 모든 체결 정보를 담은 DataFrame

        - chetime (datetime.time): 체결시간
        - sign (int): 전일대비구분 (1:상한 2:상승 3:보함 4:하한 5:하락)
        - change (int): 전일대비가격
        - drate (float): 등락율
        - price (int): 체결가
        - opentime (datetime.time): 시가시간
        - open (int): 시가
        - hightime (datetime.time): 고가시간
        - high (int): 고가
        - lowtime (datetime.time): 저가시간
        - low (int): 저가
        - cgubun (str): 체결구분 (-:매도 +:매수 NaN:동시호가)
        - cvolume (int): 체결량
        - volume (int): 누적거래량
        - value (int): 누적거래대금
        - mdvolume (int): 매도체결수량
        - msvolume (int): 매수체결수량
        - mdchecnt (int): 매도체결건수
        - mschecnt (int): 매수체결건수
        - cpower (float): 체결강도
        - w_avrg (int): 가중평균가
        - offerho (int): 매도호가
        - bidho (int): 매수호가
        - status (int): 장정보 (0:장중 10:장전시간외 4:장후시간외 3:장마감 47:시간외단일가)
        - jnilvolume (int): 전일거래량
        - shcode (str): 종목코드

    Examples:
        >>> df = get_tick_data(datetime.date.today() - datetime.timedelta(days=3), "017670")
        >>> print(df.head())
           mdchecnt  sign  mschecnt  mdvolume  w_avrg  cpower  offerho  cvolume  high  bidho  low  price cgubun  value  change  shcode   chetime opentime lowtime  volume  drate hightime  jnilvolume  msvolume  open  status
        0         1     3         0         1   37716    0.00        0        1     0      0    0  37715      -      0       0   69500  08:31:27      NaT     NaT       1    0.0      NaT           0         0     0      10
        1         2     3         0       352   37716    0.00        0      351     0      0    0  37715      -     13       0   69500  08:31:45      NaT     NaT     352    0.0      NaT           0         0     0      10
        2         2     3         1       352   37716    0.28        0        1     0      0    0  37715      +     13       0   69500  08:31:45      NaT     NaT     353    0.0      NaT           0         1     0      10
        3         2     3         2       352   37716    1.99        0        6     0      0    0  37715      +     14       0   69500  08:32:12      NaT     NaT     359    0.0      NaT           0         7     0      10
        4         2     3         3       352   37716   77.84        0      267     0      0    0  37715      +     24       0   69500  08:35:57      NaT     NaT     626    0.0      NaT           0       274     0      10

    """

    url = f"{c.PYQQQ_API_URL}/domestic-stock/daily/trades/{date}/{asset_code}"
    r = send_request(
        "GET",
        url,
    )
    if r.status_code != 200 and r.status_code != 201:
        logger.error(f"Failed to get tick data: {r.status_code}")
        return None

    with BytesIO() as byte_stream:
        byte_stream.write(r.content)
        byte_stream.seek(0)

        if date < dtm.date(2024, 4, 19):
            dtype = {"shcode": "str"}
            df = pd.read_csv(byte_stream, compression="xz", dtype=dtype)

            for col in ["chetime", "opentime", "hightime", "lowtime"]:
                df[col] = pd.to_datetime(df[col], format="%H%M%S").dt.time
        else:
            dtype = {"code": "str"}
            df = pd.read_csv(byte_stream, compression="xz", dtype=dtype)

            df["time"] = pd.to_datetime(df["time"], format="%H%M%S").dt.time

    return df
